@@ -1,7 +1,7 @@
 /*
  * pg_bulkload: lib/writer_direct.c
  *
- *	  Copyright (c) 2007-2010, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ *	  Copyright (c) 2007-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 #include "pg_bulkload.h"
@@ -83,7 +83,6 @@ static void	flush_pages(DirectWriter *loader);
 static void	close_data_file(DirectWriter *loader);
 static void	UpdateLSF(DirectWriter *loader, BlockNumber num);
 static void UnlinkLSF(DirectWriter *loader);
-static void ValidateLSFDirectory(const char *path);
 
 /* ========================================================================
  * Implementation
@@ -111,7 +110,7 @@ CreateDirectWriter(Oid relid, const WriterOptions *options)
 	self->curblk = 0;
 
 	self->rel = heap_open(relid, AccessExclusiveLock);
-	VerifyTarget(self->rel);
+	VerifyTarget(self->rel, options->max_dup_errors);
 
 	SpoolerOpen(&self->spooler, self->rel, false, options);
 	self->base.context = GetPerTupleMemoryContext(self->spooler.estate);
@@ -508,7 +507,7 @@ UnlinkLSF(DirectWriter *loader)
 /*
  * Check for LSF directory. If not exists, create it.
  */
-static void
+void
 ValidateLSFDirectory(const char *path)
 {
 	struct stat	stat_buf;
